import json
from handwriting_sample.base import LoggableObject


# ------------ #
# File writers #
# ------------ #

class JSONFileWriter(LoggableObject):
    """Class implementing JSON file writer"""

    @classmethod
    def write(cls, path, data, meta=None, verbose=False):
        """Writes the handwriting data and meta data to a JSON file"""
        try:
            with open(path, "w") as f:
                json.dump({"meta_data": meta, "data": data}, f)
                cls.log(f"Data stored in a JSON file: {path}", be_verbose=verbose)
                return True
        except Exception as e:
            cls.log(f"Unable to store to a JSON file: {path} due to {e}")
            raise


class SVCFileWriter(LoggableObject):
    """Class implementing SVC file writer"""

    @classmethod
    def write(cls, path, data, meta=None, verbose=False):
        """Writes the handwriting data and meta data to an SVC file"""
        try:
            with open(path, "w") as f:
                f.writelines(f"{meta.get('samples_count')}\n")
                f.writelines(data)
                cls.log(f"Data stored in an SVC file: {path}", be_verbose=verbose)
        except Exception as e:
            cls.log(f"Unable to store to an SVC file: {path} due to {e}")
            raise



from handwriting_sample.base import HandwritingDataBase
from handwriting_sample.validator.exceptions import PenStatusException, NegativeValueException


class HandwritingSampleValidator(HandwritingDataBase):
    """Class implementing handwriting data validator"""

    # ------------------ #
    # Validation methods #
    # ------------------ #
    # TODO: idea: make library specific exceptions

    @classmethod
    def validate_data(cls, df_data, verbose=False):
        """Validates input data"""

        # Set column names to lower case
        df_data.columns = [x.lower() for x in df_data.columns]

        # Get column count
        columns = df_data.columns

        # Check for number of time-series
        if len(columns) < len(cls.COLUMNS):
            raise ValueError(
                f"Input data are missing the following mandatory time-series (columns): "
                f"{list(set(columns).symmetric_difference(set(cls.COLUMNS)))}")
        if len(columns) > len(cls.COLUMNS):
            raise ValueError(
                f"Input data have unwanted time-series that are not expected in the data: "
                f"{list(set(columns).symmetric_difference(set(cls.COLUMNS)))}")

        # Check for missing values:
        if df_data.isnull().sum().sum() > 0:
            raise ValueError(
                f"Empty values in input data. Please inspect your input and replace the emtpy values. \n"
                f"The following table shows the count of emtpy values in particular columns: \n"
                f"{df_data.isnull().sum()}")

        # Check if the values are numerical
        for column_name in columns:
            if not all(isinstance(x, (int, float)) for x in df_data[column_name]):
                raise ValueError(f"Datatype in time-series [\'{column_name}\'] is not numerical")

        # Order the columns based on the pre-defined order
        df_data = df_data[cls.COLUMNS]

        # Check if pen status contain only 0,1 values
        for index, value in enumerate(df_data[cls.PEN_STATUS]):
            if value not in [0, 1]:
                raise PenStatusException(value, index)

        # Check if data contains negative values
        df_negative_values = (df_data >= 0).all(0)
        negative_values_column_names = list(df_negative_values[df_negative_values == False].index.values)

        if negative_values_column_names:
            raise NegativeValueException(negative_values_column_names)

        # Remove any in-air movement on the boundaries
        cls._remove_first_in_air_data(df_data, verbose)
        cls._remove_last_in_air_data(df_data, verbose)

        # TODO: validate data range
        return df_data

    # --------------- #
    # Utility methods #
    # --------------- #

    @classmethod
    def _remove_first_in_air_data(cls, df, verbose):
        """Removes unwanted in-air movement at the beginning of writing"""
        if verbose:
            cls.log(f"Check if data contains first in-air movement (unwanted before writing)")

        # Check if the first sample has any in air movement
        if df[cls.PEN_STATUS].iloc[0] == 1:
            if verbose:
                cls.log(f"Data do not contain any in-air movement at the beginning")
            return

        # Remove in-air data at the beginning
        if verbose:
            cls.log(f"Data contains in-air movement at the beginning")

        count = 0
        for index, row in df.iterrows():
            if row[cls.PEN_STATUS] == 0:
                df.drop(index, inplace=True)
                count += 1
            else:
                if verbose:
                    cls.log(f"Removed first {count} in-air samples")
                df.reset_index(inplace=True)
                return

    @classmethod
    def _remove_last_in_air_data(cls, df, verbose):
        """Removes unwanted in-air movement at the end of writing"""
        if verbose:
            cls.log(f"Check if data contains last in-air movement (unwanted after writing)")

        # Check if last sample has any in-air movement
        if df[cls.PEN_STATUS].iloc[-1] == 1:
            if verbose:
                cls.log(f"Data do not contain any in-air movement at the end")
            return

        # Remove in-air data at the end
        if verbose:
            cls.log(f"Data contains in-air movement at the beginning")

        count = 0
        for index in range(df.shape[0] - 1, -1, -1):
            if df[cls.PEN_STATUS].iloc[index] == 0:
                df.drop(index, inplace=True)
                count += 1
            else:
                if verbose:
                    cls.log(f"Removed last {count} in-air samples")
                df.reset_index(inplace=True)
                return



from handwriting_sample.base.utils import log


class LoggableObject(object):
    """Base class for loggable objects"""

    @classmethod
    def log(cls, text, be_verbose=False, is_verbose=False):
        """
        Logs the specified text.

        More info:
        ``be_verbose`` is True if logging should be verbose
        ``is_verbose`` is True if the specified text to log is verbose

        :param text: a text to log
        :type text: str
        :param be_verbose: verbosity of the logging, defaults to False
        :type be_verbose: bool, optional
        :param is_verbose: verbosity of the text, defaults to False
        :type is_verbose: bool, optional
        :return: None
        :rtype: None type
        """
        log(text, cls.__name__, be_verbose, is_verbose)


class HandwritingDataBase(LoggableObject):
    """Base class for handwriting sample"""

    # Handwriting data
    AXIS_X = "x"
    AXIS_Y = "y"
    TIME = "time"
    PEN_STATUS = "pen_status"
    AZIMUTH = "azimuth"
    TILT = "tilt"
    PRESSURE = "pressure"

    # Columns
    COLUMNS = [AXIS_X, AXIS_Y, TIME, PEN_STATUS, AZIMUTH, TILT, PRESSURE]

    # Date formant
    DATE_FORMAT = "%Y-%m-%d, %H:%M:%S"



from handwriting_sample.base import HandwritingDataBase


class TransformerException(Exception, HandwritingDataBase):
    """ Base class for TransformerException """

    def __init__(self, message):
        super(TransformerException, self).__init__(message)
        self.log(message)


class TransformerAngleTypeException(TransformerException):
    """ Exception raised for errors in the input angle type.

       Attributes:
           angle_type -- input angle_type which caused the error
           message -- explanation of the error
    """

    def __init__(self, angle_type):
        self.angle = angle_type

        if not self.angle:
            self.message = f"Angle type has not been specified! " \
                           f"Please select from ['{self.TILT}', '{self.AZIMUTH}']."
        else:
            self.message = f"Unknown Angle Type '{self.angle}' for HandwritingSample object instance. " \
                           f"Please select from ['{self.TILT}', '{self.AZIMUTH}']."
        
        super(TransformerAngleTypeException, self).__init__(self.message)


import numpy as np
from handwriting_sample.base import HandwritingDataBase


class HandwritingSampleTransformer(HandwritingDataBase):
    """Class implementing handwriting data transformer"""

    # Define conversion variables
    INCH_TO_MM = 25.4
    CM_TO_MM = 10

    # Define conversion types
    LPI = "lpi"
    LPMM = "lpmm"
    MM = "mm"

    # Define device defaults (DTK-1660)
    LPI_VALUE = 5080
    LPMM_VALUE = 200
    MM_VALUE = 0.01
    MAX_PRESSURE_VALUE = 32767
    PRESSURE_LEVELS = 8192
    MAX_TILT_VALUE = 900
    MAX_AZIMUTH_VALUE = 3600

    MAX_TILT_DEGREE = 90
    MAX_AZIMUTH_DEGREE = 360

    MAX_OLD_RANGE_PRESSURE = 1024

    @staticmethod
    def normalize_time_series(input_array, max_value):
        """
        Normalizes input time-series.

        :param input_array: Array of input data
        :type input_array: nd.array
        :param max_value: max heoretical value
        :type max_value: int
        :return: Normalized data
        :rtype: list
        """

        # Check input
        if not all(isinstance(x.item(), (int, float)) for x in input_array):
            raise ValueError(f"Input data are not numbers!")
        if not isinstance(max_value, (int, float)):
            raise ValueError(f"Max value is not number!")

        # Return normalized array
        return np.array(list(map((lambda x: x / max_value), input_array)))

    @staticmethod
    def normalize_pressure(input_array, max_value=MAX_PRESSURE_VALUE, pressure_levels=PRESSURE_LEVELS):
        """
        Normalizes pressure to pressure level of the device.

        :param input_array: Input array with pressure values
        :type input_array: np.array
        :param max_value: OPTIONAL, DEFAULT = 32767
                          max theoretical raw pressure value
        :type max_value: int
        :param pressure_levels: OPTIONAL, DEFAULT = 8192
                                level of pressure of the device
        :type pressure_levels: int
        :return: array with normalized pressure
        :rtype: np.array
        """

        # Check input
        if not all(isinstance(x.item(), (int, float)) for x in input_array):
            raise ValueError(f"Input data are not numbers!")
        if not isinstance(max_value, (int, float)):
            raise ValueError(f"Max value is not number!")
        if not isinstance(pressure_levels, (int, float)):
            raise ValueError(f"Pressure levels is not number!")

        return np.array(list(map((lambda x: (x / max_value) * pressure_levels), input_array)))

    @staticmethod
    def transform_time_to_seconds(time_array):
        """
        Transforms time to seconds.

        :param time_array: input array of timestamp
        :type time_array: np.array
        :return: time in seconds
        :rtype: nd.array
        """

        # Check input
        if not all(isinstance(x.item(), (int, float)) for x in time_array):
            raise ValueError(f"Input data are not numbers!")

        return np.array(list(map((lambda x: (x - time_array[0]) / 1e3), time_array)))

    @staticmethod
    def transform_angle(input_array, max_raw_value, max_degree_value):
        """
        Transforms raw angle to degrees.

        :param input_array: Input array with raw angle values
        :type input_array: np.array
        :param max_raw_value: Maximal theoretical value of raw angle
        :type max_raw_value: int
        :param max_degree_value: Maximal value of angle in degrees
        :type max_degree_value: int
        :return: angle in degrees
        :rtype: nd.array
        """

        # Check input
        if not all(isinstance(x.item(), (int, float)) for x in input_array):
            raise ValueError(f"Input data are not numbers!")
        if not isinstance(max_raw_value, (int, float)):
            raise ValueError(f"Max raw value is not number!")
        if not isinstance(max_degree_value, (int, float)):
            raise ValueError(f"Max angle value is not number!")

        # Get value of degree per one point
        degree_per_point = max_degree_value / max_raw_value

        # Transform array to degrees
        return np.array(list(map((lambda x: (x * degree_per_point)), input_array)))

    def transform_axis(self, sample, conversion_type=LPI, lpi_value=LPI_VALUE, lpmm_value=LPMM_VALUE,
                       shift_to_zero=True):
        """
        Transforms X,Y axis to millimeters.

        :param sample: object of HandwritingSample class
        :type sample: handwriting_sample.HandwritingSample
        :param conversion_type: OPTIONAL ["lpi"|"lpmm"|"mm"], DEFAULT="lpi".
                                Set the capturing method used for mapping;
                                "lpi" for inch; "lpmm" for millimeters;
                                "mm" for direct to millimeters
        :type conversion_type: str
        :param lpi_value: OPTIONAL, DEFAULT = 5080
                          Set lpi value of digitizing tablet
        :type lpi_value: int
        :param lpmm_value: OPTIONAL, DEFAULT = 200
                           Set lpmm value of digitizing tablet
        :type lpmm_value: int
        :param shift_to_zero: OPTIONAL, DEFAULT = True
                              Shift axis values to start from 0,0 coordinates
        :type shift_to_zero: bool
        :return: updated object of HandwritingSample class
        :rtype: handwriting_sample.HandwritingSample
        """

        # Check input
        if not isinstance(conversion_type, str):
            raise ValueError(f"Conversion type must be string not {type(conversion_type)}.")
        if not isinstance(lpi_value, int):
            raise ValueError(f"LPI value must be int not {type(lpi_value)}.")
        if not isinstance(lpmm_value, int):
            raise ValueError(f"LPMM value must be int not {type(lpmm_value)}.")

        # Check for conversion type
        if conversion_type == self.LPI:

            self.log(f"Using {conversion_type} = {lpi_value} for axis conversion to millimeters.")

            # Convert axis
            sample.x = (sample.x * self.INCH_TO_MM) / lpi_value
            sample.y = (sample.y * self.INCH_TO_MM) / lpi_value

        # BAD FORMULA... MAKING NO SENSE!
        # elif conversion_type == self.LPMM:
        #
        #     self.log(f"Using {conversion_type} = {lpmm_value} for axis conversion to millimeters.")
        #
        #     # Convert axis
        #     sample.x = (sample.x * self.INCH_TO_MM) / lpmm_value
        #     sample.y = (sample.y * self.INCH_TO_MM) / lpmm_value
        elif conversion_type == self.LPMM:
            raise NotImplementedError(f"Do not supporting this conversion anymore, due to incorrect formula.")

        elif conversion_type == self.MM:

            self.log(f"Using {conversion_type} = {self.MM_VALUE} for axis conversion to millimeters.")

            # Convert axis
            sample.x = sample.x * self.MM_VALUE
            sample.y = sample.y * self.MM_VALUE

        else:
            raise ValueError(f"Unknown conversion type {conversion_type}")

        if shift_to_zero:
            self.log(f"Shift axis data to start from 0,0 coordinates")
            sample.x = sample.x - min(sample.x)
            sample.y = sample.y - min(sample.y)

        return sample

    def transform_all_units(
            self,
            sample,
            conversion_type=LPI,
            lpi_value=LPI_VALUE,
            lpmm_value=LPMM_VALUE,
            max_raw_azimuth=MAX_AZIMUTH_VALUE,
            max_raw_tilt=MAX_TILT_VALUE,
            max_degree_azimuth=MAX_AZIMUTH_DEGREE,
            max_degree_tilt=MAX_TILT_DEGREE,
            max_pressure=MAX_PRESSURE_VALUE,
            pressure_levels=PRESSURE_LEVELS,
            angles_to_degrees=True,
            shift_to_zero=True):
        """
        Transforms all unites of sample object:
            - transforms X,Y to millimeters.
            - transform time to seconds
            - normalize or transform to degrees angles
            - normalize pressure

        :param sample: updated object of HandwritingSample class
        :type sample: handwriting_sample.HandwritingSample
        :param conversion_type: OPTIONAL ["lpi"|"lpmm"|"mm"], DEFAULT="lpi".
                                Set the capturing method used for mapping;
                                "lpi" for inch; "lpmm" for millimeters;
                                "mm" for direct to millimeters
        :type conversion_type: str
        :param lpi_value:  OPTIONAL , DEFAULT = 5080
                           Set lpi value of digitizing tablet.
        :type lpi_value: int
        :param lpmm_value: OPTIONAL, DEFAULT = 200
                           Set lpmm value of digitizing tablet.
        :type lpmm_value: int
        :param max_raw_azimuth: OPTIONAL, DEFAULT = 3600
                                Maximum theoretical value of azimuth.
        :type max_raw_azimuth: int
        :param max_raw_tilt: OPTIONAL, DEFAULT = 900
                            Maximum theoretical value of tilt.
        :type max_raw_tilt: int
        :param max_degree_azimuth: OPTIONAL, DEFAULT = 360
                                   Maximum degree value of azimuth.
        :type max_degree_azimuth: int
        :param max_degree_tilt: OPTIONAL, DEFAULT = 90
                                Maximum degree value of tilt.
        :type max_degree_tilt: int
        :param max_pressure: OPTIONAL, DEFAULT = 32767
                             Maximum theoretical value of pressure.
        :type max_pressure: int
        :param pressure_levels: OPTIONAL, DEFAULT = 8192
                                Level of pressures of the device.
        :type pressure_levels: int
        :param angles_to_degrees: OPTIONAL, DEFAULT = True
                                  Transform angles to degrees
        :type angles_to_degrees: bool
        :param shift_to_zero: OPTIONAL, DEFAULT = True
                              Shift axis values to start from 0,0 coordinates
        :type shift_to_zero: bool
        :return: updated object of HandwritingSample class
        :rtype: handwriting_sample.HandwritingSample
        """

        # TODO _read max/range values from metadata

        sample = self.transform_axis(
            sample,
            conversion_type=conversion_type,
            lpi_value=lpi_value,
            lpmm_value=lpmm_value,
            shift_to_zero=shift_to_zero)

        # Transform time to seconds
        sample.time = self.transform_time_to_seconds(sample.time)

        # Normalize Azimuth, Tilt or transform to degree
        if angles_to_degrees:

            # Transform to degrees
            sample.azimuth = self.transform_angle(
                sample.azimuth,
                max_raw_azimuth,
                max_degree_azimuth)
            sample.tilt = self.transform_angle(
                sample.tilt,
                max_raw_tilt,
                max_degree_tilt)

        # Normalize pressure
        sample.pressure = self.normalize_pressure(
            sample.pressure,
            max_value=max_pressure,
            pressure_levels=pressure_levels)

        # Return
        return sample

    def control_for_pressure(
            self,
            input_array,
            pressure_levels=PRESSURE_LEVELS,
            max_raw_press_value=MAX_PRESSURE_VALUE,
            max_range_press=MAX_OLD_RANGE_PRESSURE):
        """
        Controls for pressure range values.

        This function is a fix pressure function in case the old driver has been used
        Old pressure range is from 0-1024
        New and correct pressure range is 0-32767

        :param input_array: input pressure data
        :type input_array: np.array
        :param pressure_levels: OPTIONAL - level of pressures of the device. DEFAULT = 8192
        :type pressure_levels: int
        :param max_raw_press_value: OPTIONAL - maximum theoretical value of pressure. DEFAULT = 32767
        :type max_raw_press_value: int
        :param max_range_press: OPTIONAL - maximum allowed pressure range of raw data. DEFAULT = 1024
        :type max_range_press: int
        :return: converted pressure to new scale
        :rtype: np.array
        """

        data_pressure_range = np.ptp(input_array)
        self.log(f"Pressure range of data is: {data_pressure_range}")

        if data_pressure_range > max_range_press:

            # Convert the pressure to lower range
            self.log(f"Maximum allowed pressure range is: {max_range_press}.")
            self.log(f"Converting pressure values with following params:")
            self.log(f"  - device max pressure level = {pressure_levels}")
            self.log(f"  - raw data max pressure value = {max_raw_press_value}")

            output = np.array(list(map(lambda x: round((x / max_raw_press_value) * max_range_press), input_array)))
            print(f"Converted Pressure range is: {np.ptp(output)}")

            return output
        return input_array

    @staticmethod
    def correct_pen_status(sample):
        """
        Corrects pen status values to binary form

        :param sample: updated object of HandwritingSample class
        :type sample: handwriting_sample.HandwritingSample
        :return: instance of HandwritingSample
        :rtype: HandwritingSample
        """

        # Correct pen status
        sample.pen_status = np.array(list(map(lambda x: 1 if x > 0 else 0, sample.pressure)))

        # Hold meta data
        meta_data = sample.meta

        # Validate sample and return
        r_sample = sample.from_pandas_dataframe(sample.data_pandas_dataframe)

        # Put back teh meta data
        r_sample.add_meta_data(meta_data)

        return r_sample

    @staticmethod
    def revert_axis(input_array, axis_max_value):
        """
        Revert axis

        :param input_array: Input array with raw angle values
        :type input_array: np.array
        :return: reverted axis
        :rtype: nd.array
        """

        # Check input
        if not all(isinstance(x.item(), (int, float)) for x in input_array):
            raise ValueError(f"Input data are not numbers!")
        if not isinstance(axis_max_value, (int, float)):
            raise ValueError(f"Axis max value is not a number!")
        if max(input_array) > axis_max_value:
            raise ValueError(f"Axis max value ({axis_max_value}) is lower than max value of the input array"
                             f" ({max(input_array)})! ")
        # Revert axis
        return np.array(list(map(lambda x: axis_max_value - x, input_array)))

    @staticmethod
    def rescale_axis(sample, rescale_coef=0.5):
        """
        Rescale axis values

        :param sample: updated object of HandwritingSample class
        :type sample: handwriting_sample.HandwritingSample
        :param rescale_coef: OPTIONAL - Coefficient of rescaling, DEFAULT = 0.5
        :type rescale_coef: float
        :return: handwriting sample with rescaled axis
        :rtype: handwriting_sample.HandwritingSample
        """

        # Check input
        if not isinstance(rescale_coef, (int, float)):
            raise ValueError(f"Coefficient of rescaling is not number!")

        sample.x = sample.x * rescale_coef
        sample.y = sample.y * rescale_coef

        return sample




from handwriting_sample.base import HandwritingDataBase


class ValidatorException(Exception, HandwritingDataBase):
    """ Base class for ValidatorException """

    def __init__(self, message):
        super(ValidatorException, self).__init__(message)
        self.log(message)


class PenStatusException(ValidatorException):

    def __init__(self, value, index):
        self.message = f"Pen status column got and unexpected value on input: {value} on index {index}"

        super(PenStatusException, self).__init__(self.message)


class NegativeValueException(ValidatorException):

    def __init__(self, column_names):
        self.message = f"Input data contains negative values in the following columns: {column_names}."

        super(NegativeValueException, self).__init__(self.message)


import os
import json
import pandas as pd
from handwriting_sample.base import LoggableObject


# ------------ #
# File readers #
# ------------ #

class JSONFileReader(LoggableObject):
    """Class implementing JSON file reader"""

    @classmethod
    def read(cls, path, verbose=False):
        """Reads the handwriting data and meta data"""

        # Read the handwriting data from a JSON file
        with open(path, "r") as file:
            json_data = json.load(file)

        # Get data and meta data
        data = json_data.get("data")
        meta = json_data.get("meta_data")
        cls.log(f"Data has been loaded from a JSON file: {path}", be_verbose=verbose)

        # Return data and meta data
        return data, meta


class SVCFileReader(LoggableObject):
    """Class implementing SVC file reader"""

    @classmethod
    def read(cls, path, column_names, verbose=False):
        """Reads the handwriting data and meta data"""

        # Read the handwriting data from an SVC file
        data = pd.read_csv(path, sep=" ", names=column_names, skiprows=1).to_dict(orient="list")
        meta = cls._read_metadata_from_svc_file_name(path)
        cls.log(f"Data has been loaded from an SVC file: {path}", be_verbose=verbose)

        # Return data and meta data
        return data, meta

    @classmethod
    def _read_metadata_from_svc_file_name(cls, file_path):
        """Reads meta data included in the file name"""

        # Prepare meta data
        meta_data = {}

        # Open file and read the first line
        with open(file_path) as f:
            raw_meta_data = f.readline()

        # Store the samples count
        meta_data["samples_count"] = int(raw_meta_data)

        # Get only file name and split it to get meta data from it
        file_path = os.path.basename(os.path.splitext(file_path)[0])
        meta_from_file_name = file_path.split("_")

        # Handle two optional information included in file name for HandAQUS
        if len(meta_from_file_name) >= 4:

            meta_data["participant"] = {"id": meta_from_file_name[0]}
            meta_data["created_on"] = meta_from_file_name[-1]
            meta_data["administrator"] = meta_from_file_name[-2]
            meta_data["task_id"] = meta_from_file_name[-3]

            if len(meta_from_file_name) == 6:
                meta_data["participant"] = {
                    "id": meta_from_file_name[0],
                    "birth_date": meta_from_file_name[1],
                    "sex": meta_from_file_name[2]
                }
        else:
            cls.log("Old file-name format no additional meta data")

        # Return meta data
        return meta_data


# ------------ #
# Data readers #
# ------------ #

class ListReader(LoggableObject):
    """Class implementing list object reader"""

    @classmethod
    def read(cls, data, column_names, verbose=False):
        """Reads the handwriting data and meta data"""

        # Get the handwriting data from a list object
        data = {column: value for column, value in zip(column_names, data)}
        meta = {}
        cls.log(f"Data has been loaded from a list", be_verbose=verbose)

        # Return data and meta data
        return data, meta


class NumpyArrayReader(LoggableObject):
    """Class implementing numpy array reader"""

    @classmethod
    def read(cls, data, column_names, verbose=False):
        """Reads the handwriting data and meta data"""

        # Get the handwriting data from a list object
        data = {column: data[..., i] for i, column in enumerate(column_names)}
        meta = {}
        cls.log(f"Data has been loaded from a numpy array", be_verbose=verbose)

        # Return data and meta data
        return data, meta


class PandasDataFrameReader(LoggableObject):
    """Class implementing pandas dataframe reader"""

    @classmethod
    def read(cls, data, verbose=False):
        """Reads the handwriting data and meta data"""

        # Get the handwriting data from a list object
        data = data.to_dict(orient="list")
        meta = {}
        cls.log(f"Data has been loaded from a pandas dataframe", be_verbose=verbose)

        # Return data and meta data
        return data, meta



from handwriting_sample.base import LoggableObject
from handwriting_sample.reader.readers import (
    JSONFileReader,
    SVCFileReader,
    ListReader,
    NumpyArrayReader,
    PandasDataFrameReader
)


class HandwritingSampleReader(LoggableObject):
    """Class implementing handwriting data reader"""

    # --------------- #
    # Reading methods #
    # --------------- #

    @classmethod
    def read_from_json(cls, path, columns, verbose=False):
        """
        Reads handwriting data and meta data from a JSON file.

        :param path: path to a JSON file
        :type path: str
        :param columns: handwriting variables to be present in the data
        :type columns: list
        :param verbose: verbosity of the logging, defaults to False
        :type verbose: bool, optional
        :return: data and meta data
        :rtype: tuple
        """
        return JSONFileReader.read(path, verbose=verbose)

    @classmethod
    def read_from_svc(cls, path, columns, verbose=False):
        """
        Reads handwriting data and meta data from an SVC file.

        :param path: path to an SVC file
        :type path: str
        :param columns: handwriting variables to be present in the data
        :type columns: list
        :param verbose: verbosity of the logging, defaults to False
        :type verbose: bool, optional
        :return: data and meta data
        :rtype: tuple
        """
        return SVCFileReader.read(path, columns, verbose=verbose)

    @classmethod
    def read_from_list(cls, data, columns, verbose=False):
        """
        Reads handwriting data and meta data from a list.

        :param data: data representing handwriting sample
        :type data: list
        :param columns: handwriting variables to be present in the data
        :type columns: list
        :param verbose: verbosity of the logging, defaults to False
        :type verbose: bool, optional
        :return: data and meta data
        :rtype: tuple
        """
        return ListReader.read(data, columns, verbose=verbose)

    @classmethod
    def read_from_numpy_array(cls, data, columns, verbose=False):
        """
        Reads handwriting data and meta data from a numpy array.

        :param data: data representing handwriting sample
        :type data: np.ndarray
        :param columns: handwriting variables to be present in the data
        :type columns: list
        :param verbose: verbosity of the logging, defaults to False
        :type verbose: bool, optional
        :return: data and meta data
        :rtype: tuple
        """
        return NumpyArrayReader.read(data, columns, verbose=verbose)

    @classmethod
    def read_from_pandas_dataframe(cls, data, columns, verbose=False):
        """
        Reads handwriting data and meta data from a pandas DataFrame.

        :param data: data representing handwriting sample
        :type data: pd.DataFrame
        :param columns: handwriting variables to be present in the data
        :type columns: list
        :param verbose: verbosity of the logging, defaults to False
        :type verbose: bool, optional
        :return: data and meta data
        :rtype: tuple
        """
        return PandasDataFrameReader.read(data, verbose=verbose)



import matplotlib.pyplot as plt
from handwriting_sample.base import HandwritingDataBase


# TODO: First fast version, need to be refactored to use plotly


class HandwritingSampleVisualizer(HandwritingDataBase):
    """Class implementing handwriting data visualizer"""

    # ------------- #
    # Plot methods  #
    # ------------- #
    def plot_on_surface_movement(self, sample, x_label=None, y_label=None, save_as=None):
        """
        Plot task on surface only

        :param sample: instance of handwriting sample
        :type sample: HandwritingSample
        :param x_label: label of X axis
        :param y_label: label of Y axis
        :param save_as: setup save location if you want to save a file
        """

        # Plot task on surface only without line connections between in_air gaps
        ax, plot = self._plot_strokes(sample,
                                      x_label=x_label,
                                      y_label=y_label,
                                      on_surface_movement=True,
                                      on_surface_color='darkblue',
                                      save_as=save_as)
        return ax, plot

    def plot_in_air_movement(self, sample, x_label=None, y_label=None, save_as=None):
        """
        Plot task with in air movement

        :param sample: instance of handwriting sample
        :type sample: HandwritingSample
        :param x_label: label of X axis
        :param y_label: label of Y axis
        :param save_as: setup save location if you want to save a file
        """
        # Plot task with in air data
        ax, plot = self._plot_strokes(sample,
                                      x_label=x_label,
                                      y_label=y_label,
                                      in_air_movement=True,
                                      in_air_color='plum',
                                      save_as=save_as)

        return ax, plot

    def plot_separate_movements(self, sample, x_label=None, y_label=None, save_as=None):
        """
        Plot task with separate movements

        :param sample: instance of handwriting sample
        :type sample: HandwritingSample
        :param x_label: label of X axis
        :param y_label: label of Y axis
        :param save_as: setup save location if you want to save a file
        """
        # Plot task with in air data
        ax, plot = self._plot_strokes(sample,
                                      x_label=x_label,
                                      y_label=y_label,
                                      in_air_movement=True,
                                      on_surface_movement=True,
                                      on_surface_color='darkblue',
                                      in_air_color='plum',
                                      save_as=save_as)

        return ax, plot

    def plot_strokes(self, sample, x_label=None, y_label=None, save_as=None):
        """
        Plot task separate strokes

        :param sample: instance of handwriting sample
        :type sample: HandwritingSample
        :param x_label: label of X axis
        :param y_label: label of Y axis
        :param save_as: setup save location if you want to save a file
        """
        # Plot task with in air data
        ax, plot = self._plot_strokes(sample,
                                      x_label=x_label,
                                      y_label=y_label,
                                      in_air_movement=True,
                                      on_surface_movement=True,
                                      save_as=save_as)

        return ax, plot

    def _plot_strokes(
            self,
            sample,
            x_label=None,
            y_label=None,
            in_air_movement=False,
            on_surface_movement=False,
            on_surface_color=None,
            in_air_color=None,
            fig_show=True,
            save_as=None,
            fig_kwargs=None):
        """
        Plot line plots for input data

        :param sample: instance of handwriting sample
        :type sample: handwriting_sample.HandwritingSample
        :param x_label: label of X axis
        :param y_label: label of Y axis
        :param in_air_movement: show in air movements
        :param on_surface_movement: show on surface movements
        :param fig_show: True/False value to show the figure
        :param save_as: setup save location if you want to save a file
        :return: axis and figure of the plot
        """

        # Prepare the figure settings
        fig_kwargs = fig_kwargs if fig_kwargs else {
            "fig_size": (10, 6),
            "show_ticks": True,
            "x_label": "X [mm]",
            "y_label": "Y [mm]",
            "title": "Strokes XY",
            "shade": True,
            "colors": ["red", "blue"]
        }

        # Get strokes
        stroke_data = sample.get_strokes(on_surface_only=on_surface_movement,
                                         in_air_only=in_air_movement)

        if not stroke_data:
            self.log(f"No data to plot!")
            return

        # Create figure if axes not inserted
        fig, axis = plt.subplots(1, 1, figsize=fig_kwargs.get("fig_size"))

        # Set labels
        x_label = fig_kwargs["x_label"] if not x_label else x_label
        y_label = fig_kwargs["y_label"] if not y_label else y_label

        # Plot strokes
        for stroke in stroke_data:
            self._plot_task(stroke,
                            x_label=x_label,
                            y_label=y_label,
                            ax=axis,
                            on_surface_color=on_surface_color,
                            in_air_color=in_air_color)

        # Set tight layout
        fig.tight_layout()

        # Save the graph
        if save_as:
            plt.savefig(save_as, bbox_inches="tight")

        # Show the graph
        if fig_show:
            plt.show()
        else:
            plt.close()

        return axis, plt

    @staticmethod
    def _plot_task(
            stroke,
            x_label=None,
            y_label=None,
            on_surface_color=None,
            in_air_color=None,
            ax=None,
            fig=None,
            fig_show=True,
            save_as=None,
            fig_kwargs=None):
        """
        Plot line plots for input data

        :param stroke: instance of handwriting sample
        :param x_label:           label of X axis
        :param y_label:           label of Y axis
        :param on_surface_color:  color of on surface movement
        :param in_air_color:      color of in air movement
        :param ax:                axis to plot on
        :param fig:               figure to show on
        :param fig_show:          True/False value to show the figure
        :param save_as:           setup save location if you want to save a file

        :return:                  axis and figure of the plot
        """

        # Prepare the figure settings
        fig_kwargs = fig_kwargs if fig_kwargs else {
            "fig_size": (10, 6),
            "show_ticks": True,
            "x_label": "X [mm]",
            "y_label": "Y [mm]",
            "title": "Task XY",
            "shade": False,
            "colors": ["red", "blue"]
        }

        # Split movements to on surface and in air
        # on_surface, in_air = split_movement(data)

        # Create figure if axes not inserted
        if ax is None:
            fig, ax = plt.subplots(1, 1, figsize=fig_kwargs.get("fig_size"))

        # Plot on surface
        color = on_surface_color if stroke[0] == "on_surface" else in_air_color
        ax.plot(stroke[1].x, stroke[1].y, color=color)

        # Plot in air if True
        # if in_air_movement:
        #     ax.plot(in_air['X'], in_air['Y'], color=in_air_color)

        # Set Grid to ture
        ax.yaxis.grid(True)
        ax.xaxis.grid(True)

        # Set labels
        title = fig_kwargs.get("title")
        x_label = fig_kwargs.get("x_label") if not x_label else x_label
        y_label = fig_kwargs.get("y_label") if not y_label else y_label

        ax.set_xlabel(x_label if x_label else "", fontname="Times New Roman", fontsize=16)
        ax.set_ylabel(y_label if y_label else "", fontname="Times New Roman", fontsize=16)

        # If figure is created set layout and store if wanted
        if fig:
            fig.tight_layout()

            # Save the graph
            if save_as:
                plt.savefig(save_as, bbox_inches="tight")

            # Show the graph
            if fig_show:
                plt.show()
            else:
                plt.close()

        # Return
        return ax, plt

    def plot_all_modalities(self, sample, x_label=None, save_as=None):
        """
        Plot separate line plots with all modalities of data

        :param sample: instance of handwriting sample
        :type sample: handwriting_sample.HandwritingSample
        :param x_label: label of X axis
        :param save_as: setup save location if you want to save a file
        """

        # Go over modalities and plot them
        df_data = sample.data_pandas_dataframe
        for col in df_data:
            # Set labels
            x_label = "sample" if not x_label else x_label
            y_label = col

            # Plot lines
            self.plot_line(df_data[col], x_label, y_label, save_as=save_as)

    @staticmethod
    def plot_line(data, x_label, y_label, fig_show=True, save_as=None, fig_kwargs=None):
        """
        Plot line plots for input data

        :param data: input data array
        :param x_label: label of X axis
        :param y_label: label of Y axis
        :param fig_show: True/False value to show the figure
        :param save_as: setup save location if you want to save a file
        :return: axis and figure of the plot
        """

        # Prepare the figure settings
        fig_kwargs = fig_kwargs if fig_kwargs else {
            "fig_size": (15, 4),
            "show_ticks": True,
            "x_label": x_label,
            "y_label": y_label,
            "shade": True,
            "title": f"{y_label}",
            "colors": ["red", "blue", "green"]
        }

        # Create figure if axes not inserted
        fig, ax = plt.subplots(1, 1, figsize=fig_kwargs.get("fig_size"))

        # Plot
        ax.plot(data, linewidth=2, alpha=0.7)

        # Set labels
        title = fig_kwargs.get("title")
        x_label = fig_kwargs.get("x_label")
        y_label = fig_kwargs.get("y_label")

        ax.set_title(title if title else "", fontname="Times New Roman", fontsize=12)
        ax.set_xlabel(x_label if x_label else "", fontname="Times New Roman", fontsize=16)
        ax.set_ylabel(y_label if y_label else "", fontname="Times New Roman", fontsize=16)

        # Set grid
        ax.yaxis.grid(True)
        ax.xaxis.grid(True)

        # Set tight layout
        fig.tight_layout()

        # Save the graph
        if save_as:
            plt.savefig(save_as, bbox_inches="tight", quality=100)

        # Show the graph
        if fig_show:
            plt.show()
        else:
            plt.close()

        return ax, plt



import os
from datetime import datetime
from handwriting_sample.base import LoggableObject
from handwriting_sample.writer.writers import JSONFileWriter, SVCFileWriter


class HandwritingSampleWriter(LoggableObject):
    """Class implementing handwriting data writer"""

    # --------------- #
    # Writing methods #
    # --------------- #
    # TODO: idea: there is some common functionality in store_... methods that may be taken out into a common method

    def write_to_json(self, sample, save_path, file_name=None, store_original_data=False, verbose=False):
        """
        Stores HandwritingSample data to a JSON file.

        :param sample: instance of handwriting sample
        :type sample: HandwritingSample
        :param save_path: save path
        :type save_path: str
        :param file_name: file name (optional if meta data), defaults to None
        :type file_name: str, optional
        :param store_original_data: store original data, defaults to False
        :type store_original_data: bool, optional
        :param verbose: verbosity of the logging, defaults to False
        :type verbose: bool, optional
        :return: True if stored, False otherwise
        :rtype: bool
        """

        # Get the data and meta data from the handwriting sample
        data = sample.original_data_pandas_dataframe if store_original_data else sample.data_pandas_dataframe
        meta = sample.meta

        # Prepare and validate the data and meta data
        data = sample.validator.validate_data(data, verbose=verbose)
        meta = self._prepare_meta_data(sample, meta)

        # If the filename is not set, create a default one
        if not file_name and meta:
            file_name = self._collect_file_name(meta)

        # Update the save_path
        save_path = os.path.join(save_path, f"{file_name}.json")

        # Prepare the data to be stored
        data = data.to_dict("list")

        # Write the data
        return JSONFileWriter.write(save_path, data, meta=meta, verbose=verbose)

    def write_to_svc(self, sample, save_path, file_name=None, store_original_data=False, verbose=False):
        """
        Stores HandwritingSample data to an SVC file.

        :param sample: instance of handwriting sample
        :type sample: HandwritingSample
        :param save_path: save path
        :type save_path: str
        :param file_name: file name (optional if meta data), defaults to None
        :type file_name: str, optional
        :param store_original_data: store original data, defaults to False
        :type store_original_data: bool, optional
        :param verbose: verbosity of the logging, defaults to False
        :type verbose: bool, optional
        :return: True if stored, False otherwise
        :rtype: bool
        """

        # Get the data and meta data from the handwriting sample
        data = sample.original_data_pandas_dataframe if store_original_data else sample.data_pandas_dataframe
        meta = sample.meta

        # Prepare and validate the data and meta data
        data = sample.validator.validate_data(data)
        meta = self._prepare_meta_data(sample, meta)

        # If the filename is not set, create a default one
        if not file_name and meta:
            file_name = self._collect_file_name(meta)

        # Update the save_path
        save_path = os.path.join(save_path, f"{file_name}.svc")

        # Prepare the data to be stored
        data = data.to_dict("r")
        data = [
            f"{row[sample.AXIS_X]} {row[sample.AXIS_Y]} "
            f"{row[sample.TIME]} {row[sample.PEN_STATUS]} "
            f"{row[sample.AZIMUTH]} {row[sample.TILT]} "
            f"{row[sample.PRESSURE]}\n" for row in data
        ]

        # Write the data
        return SVCFileWriter.write(save_path, data, meta=meta, verbose=verbose)

    # --------------- #
    # Utility methods #
    # --------------- #

    @classmethod
    def _prepare_meta_data(cls, sample, meta_data=None):
        """Prepares the meta data before writing"""

        # Handle no meta data
        meta_data = meta_data if meta_data else {}

        # Create or update the timestamps
        meta_data["written_on"] = datetime.utcnow().strftime(sample.DATE_FORMAT)
        meta_data["created_on"] = datetime.utcnow().strftime(sample.DATE_FORMAT) \
            if not meta_data.get("created_on", None) \
            else meta_data.get("created_on", None)

        # Return the prepared meta data
        return meta_data

    @classmethod
    def _collect_file_name(cls, meta_data):
        """Collects information from meta data and creates a filename"""

        # Handle "old" data
        if meta_data.get("participant", None) is None:
            raise ValueError(f"No proper meta data for this sample, please select filename manually")

        # Collect the information
        subject_info = {
            "participant_id": meta_data.get("participant", {}).get("id", None),
            "birth_date": meta_data.get("participant", {}).get("birth_date", None),
            "sex": meta_data.get("participant", {}).get("sex", None),
            "task_id": meta_data.get("task_id", None),
            "admin": meta_data.get("administrator", None),
            "created_on": meta_data.get("created_on", None)
        }

        # Collect the filename
        file_name = ""
        last_value = list(subject_info.values())[-1]

        for value in subject_info.values():
            if value:
                file_name += f"{value}_" if value not in last_value else f"{value}"

        # Return the filename
        return file_name



import numpy as np
import pandas as pd
from datetime import datetime
from handwriting_sample.base import HandwritingDataBase
from handwriting_sample.reader import HandwritingSampleReader
from handwriting_sample.writer import HandwritingSampleWriter
from handwriting_sample.validator import HandwritingSampleValidator
from handwriting_sample.transformer import HandwritingSampleTransformer, TransformerAngleTypeException
from handwriting_sample.visualizer import HandwritingSampleVisualizer


class HandwritingSample(HandwritingDataBase):
    """Class implementing the management of sample handwriting samples"""

    # Handwriting data helpers (reading, writing, validation, transformer, visualizer)
    reader = HandwritingSampleReader()
    writer = HandwritingSampleWriter()
    validator = HandwritingSampleValidator()
    transformer = HandwritingSampleTransformer()
    visualizer = HandwritingSampleVisualizer()

    # TODO: idea: I think np.column_stack is going to work if X, Y, etc. are 1D numpy arrays as well
    def __init__(self, x, y, time, pen_status, azimuth, tilt, pressure, meta_data=None, validate=True, verbose=False):
        """
        Initializes the HandwritingSample object.

        :param x: X axis
        :type x: list[uint]
        :param y: Y axis
        :type y: list[uint]
        :param time: timestamp
        :type time: list[uint]
        :param pen_status: indication of pen location (on-surface=1 | in-air=0)
        :type pen_status: list[bool]
        :param azimuth: azimuth of the pen
        :type azimuth: list[uint]
        :param tilt: tilt of the pen
        :type tilt: list[uint]
        :param pressure: pressure value
        :type pressure: list[uint]
        :param meta_data: dictionary with meta data
        :type meta_data: dict
        :param validate: true if validate input data
        :type validate:bool
        :param verbose: true if log should be verbose
        :type verbose: bool
        """

        # Create pandas DataFrame object from the input handwriting variables
        df = pd.DataFrame(np.column_stack([x, y, time, pen_status, azimuth, tilt, pressure]), columns=self.COLUMNS)

        # Validate and store input data
        self._data = self.validator.validate_data(df, verbose=verbose) if validate else df

        # Store meta data of any kind
        self.meta = meta_data

        # Set the handwriting variables
        self.x = self._data[self.AXIS_X].to_numpy()
        self.y = self._data[self.AXIS_Y].to_numpy()
        self.time = self._data[self.TIME].to_numpy()
        self.pen_status = self._data[self.PEN_STATUS].to_numpy(dtype=bool if validate else None)
        self.azimuth = self._data[self.AZIMUTH].to_numpy()
        self.tilt = self._data[self.TILT].to_numpy()
        self.pressure = self._data[self.PRESSURE].to_numpy()

    def __repr__(self):
        return f"<HandwritingSampleObject: \n" \
               f"DATA:\n" \
               f"   x =          {self.x}, \n" \
               f"   y =          {self.y}, \n" \
               f"   time =       {self.time}, \n" \
               f"   pen_status = {self.pen_status}, \n" \
               f"   azimuth =    {self.azimuth}, \n" \
               f"   tilt =       {self.tilt}, \n" \
               f"   pressure =   {self.pressure}> \n\n\n" \
               f"METADATA:\n" \
               f"{self.meta.items() if self.meta else None}"

    # ---------- #
    # Properties #
    # ---------- #

    @property
    def data_list(self):
        """Returns list for the non-original data"""
        return [self.x, self.y, self.time, self.pen_status, self.azimuth, self.tilt, self.pressure]

    @property
    def data_numpy_array(self):
        """Returns numpy array for the non-original data"""
        return np.column_stack(self.data_list)

    @property
    def data_pandas_dataframe(self):
        """Returns pandas DataFrame for the non-original data"""
        return pd.DataFrame(self.data_numpy_array, columns=self.COLUMNS)

    @property
    def original_data_list(self):
        """Returns list for the original data"""
        return [self._data[column].to_numpy() for column in self._data.columns]

    @property
    def original_numpy_array(self):
        """Returns numpy array for the original data"""
        return self._data.values

    @property
    def original_data_pandas_dataframe(self):
        """Returns pandas DataFrame for the original data"""
        return self._data

    @property
    def xy(self):
        """Returns general movement of X and Y"""
        return np.sqrt(np.power(self.x, 2) + np.power(self.y, 2))

    # --------------- #
    # Reading methods #
    # --------------- #

    @classmethod
    def from_json(cls, path, columns=None, validate=True):
        """
        Creates a HandwritingSample instance from a JSON file.

        :param path: path to a JSON file
        :type path: str
        :param columns: handwriting variables, defaults to cls.COLUMNS
        :type columns: list, optional
        :param validate: true if validate input data
        :type validate:bool
        :return: instance of HandwritingSample
        :rtype: HandwritingSample
        """
        return cls._from_data_and_metadata(*cls.reader.read_from_json(path, columns or cls.COLUMNS), validate=validate)

    @classmethod
    def from_svc(cls, path, columns=None, validate=True):
        """
        Creates a HandwritingSample instance from an SVC file.

        :param path: path to an SVC file
        :type path: str
        :param columns: handwriting variables, defaults to cls.COLUMNS
        :type columns: list, optional
        :param validate: true if validate input data
        :type validate:bool
        :return: instance of HandwritingSample
        :rtype: HandwritingSample
        """
        return cls._from_data_and_metadata(*cls.reader.read_from_svc(path, columns or cls.COLUMNS), validate=validate)

    @classmethod
    def from_list(cls, data, columns=None, validate=True):
        """
        Creates a HandwritingSample instance from a list.

        :param data: data representing handwriting sample
        :type data: list
        :param columns: handwriting variables, defaults to cls.COLUMNS
        :type columns: list, optional
        :param validate: true if validate input data
        :type validate:bool
        :return: instance of HandwritingSample
        :rtype: HandwritingSample
        """
        return cls._from_data_and_metadata(*cls.reader.read_from_list(data, columns or cls.COLUMNS), validate=validate)

    @classmethod
    def from_numpy_array(cls, data, columns=None, validate=True):
        """
        Creates a HandwritingSample instance from a numpy array.

        :param data: data representing handwriting sample
        :type data: np.ndarray
        :param columns: handwriting variables, defaults to cls.COLUMNS
        :type columns: list, optional
        :param validate: true if validate input data
        :type validate:bool
        :return: instance of HandwritingSample
        :rtype: HandwritingSample
        """
        return cls._from_data_and_metadata(*cls.reader.read_from_numpy_array(data, columns or cls.COLUMNS), validate=validate)

    @classmethod
    def from_pandas_dataframe(cls, data, columns=None, validate=True):
        """
        Creates a HandwritingSample instance from a pandas DataFrame.

        :param data: data representing handwriting sample
        :type data: pd.DataFrame
        :param columns: handwriting variables, defaults to cls.COLUMNS
        :type columns: list, optional
        :param validate: true if validate input data
        :type validate:bool
        :return: instance of HandwritingSample
        :rtype: HandwritingSample
        """
        return cls._from_data_and_metadata(*cls.reader.read_from_pandas_dataframe(data, columns or cls.COLUMNS), validate=validate)

    @classmethod
    def _from_data_and_metadata(cls, data, meta_data=None, validate=True):
        """
        Creates a HandwritingSample instance from data and meta data.

        :param data: data of the handwriting sample
        :type data: dict
        :param meta_data: meta data of the handwriting sample, defaults to None
        :type meta_data: dict, optional
        :param validate: true if validate input data
        :type validate:bool
        :return: instance of HandwritingSample
        :rtype: HandwritingSample
        """
        return cls(**data, meta_data=meta_data or {}, validate=validate)

    # --------------- #
    # Writing methods #
    # --------------- #

    def to_json(self, path, file_name=None, store_original_data=False):
        """
        Writes sample data to a JSON file.

        :param path: path where data should be stored
        :type path: str
        :param file_name: custom file name, defaults to None
        :type file_name: str, optional
        :param store_original_data: store original data, defaults to False
        :type store_original_data: bool, optional
        :return: None
        :rtype: None type
        """
        return self.writer.write_to_json(self, path, file_name=file_name, store_original_data=store_original_data)

    def to_svc(self, path, file_name=None, store_original_data=False):
        """
        Writes sample data to an SVC file.

        :param path: path where data should be stored
        :type path: str
        :param file_name: custom file name, defaults to None
        :type file_name: str, optional
        :param store_original_data: store original data, defaults to False
        :type store_original_data: bool, optional
        :return: None
        :rtype: None type
        """
        return self.writer.write_to_svc(self, path, file_name=file_name, store_original_data=store_original_data)

    # ----------------------------- #
    # Handwriting data manipulation #
    # ----------------------------- #

    def get_on_surface_data(self):
        """Returns on-surface data as a HandwritingSample object"""

        # Get all on-surface data
        df = self.data_pandas_dataframe
        df = df[df[self.PEN_STATUS] == 1]

        # Return a new instance of HandwritingSample with only on-surface data
        return HandwritingSample(**df.to_dict(orient="list"), validate=False)

    def get_in_air_data(self):
        """Returns in-air data as a HandwritingSample object"""

        # Return all in-air data
        df = self.data_pandas_dataframe
        df = df[df[self.PEN_STATUS] == 0]

        # Return a new instance of HandwritingSample with only in-air data
        return HandwritingSample(**df.to_dict(orient="list"), validate=False)

    def get_on_surface_strokes(self):
        """Returns strokes on-surface"""
        return self.get_strokes(on_surface_only=True)

    def get_in_air_strokes(self):
        """Returns strokes in-air"""
        return self.get_strokes(in_air_only=True)

    def get_strokes(self, on_surface_only=False, in_air_only=False):
        """
        Splits the movement into strokes.

        :param on_surface_only: on-surface strokes only, defaults to False
        :type on_surface_only: bool, optional
        :param in_air_only: in-air strokes only, defaults to True
        :type in_air_only: bool, optional
        :return: list of strokes in tuples with the status of strokes
        :rtype: tuple('status', HandwritingSample)
        """

        # Handle the edge cases
        if all((on_surface_only, in_air_only)):
            on_surface_only, in_air_only = False, False

        # Get accessible data of the sample as a pandas DataFrame
        df = self.data_pandas_dataframe

        # Get index values of the pen status column changes
        idx_change = df.ne(df.shift()).filter(like=self.PEN_STATUS).apply(lambda x: x.index[x].tolist())
        idx_array = idx_change[self.PEN_STATUS].values.tolist()

        # Add the last index value
        idx_array.append(df.index[-1])

        # Get strokes
        strokes = [df.iloc[idx_array[n]:idx_array[n + 1]] for n in range(len(idx_array) - 1)]

        # Prepare the list of strokes
        list_of_strokes = []

        # Fill the list of strokes (add 'on_surface'/'in_air' flag in front of each stroke)
        for stroke in strokes:

            # Skip empty strokes
            if stroke.empty:
                continue

            # If on surface strokes only are wanted filter out in air
            if on_surface_only and stroke[self.PEN_STATUS].iloc[0] == 0:
                continue

            # If in air strokes only are wanted filter out on surface
            if in_air_only and stroke[self.PEN_STATUS].iloc[0] == 1:
                continue

            # Prepare the stroke information
            status = "on_surface" if stroke[self.PEN_STATUS].iloc[0] == 1 else "in_air"
            stroke = HandwritingSample(**stroke.to_dict(orient="list"), validate=False)

            # Append to the list of strokes
            list_of_strokes.append((status, stroke))

        # Return the list of strokes
        return list_of_strokes

    # ------------------------------- #
    # Handwriting data transformation #
    # ------------------------------- #
    # TODO: use **kwargs
    def transform_all_units(
            self,
            conversion_type=transformer.LPI,
            lpi_value=transformer.LPI_VALUE,
            lpmm_value=transformer.LPMM_VALUE,
            max_raw_azimuth=transformer.MAX_AZIMUTH_VALUE,
            max_raw_tilt=transformer.MAX_TILT_VALUE,
            max_degree_azimuth=transformer.MAX_AZIMUTH_DEGREE,
            max_degree_tilt=transformer.MAX_TILT_DEGREE,
            max_pressure=transformer.MAX_PRESSURE_VALUE,
            pressure_levels=transformer.PRESSURE_LEVELS,
            angles_to_degrees=True,
            shift_to_zero=True):
        """
        Transforms all unites of sample object:
            - transforms X,Y to millimeters.
            - transform time to seconds
            - normalize or transform to degrees angles
            - normalize pressure

        :param conversion_type: OPTIONAL ["lpi"|"lpmm"], DEFAULT="lpi".
                                Set the capturing method used for mapping;
                                "lpi" for inch; "lpmm" for millimeters
        :type conversion_type: str
        :param lpi_value:  OPTIONAL , DEFAULT = 5080
                           Set lpi value of digitizing tablet.
        :type lpi_value: int
        :param lpmm_value: OPTIONAL, DEFAULT = 200
                           Set lpmm value of digitizing tablet.
        :type lpmm_value: int
        :param max_raw_azimuth: OPTIONAL, DEFAULT = 3600
                                Maximum theoretical value of azimuth.
        :type max_raw_azimuth: int
        :param max_raw_tilt: OPTIONAL, DEFAULT = 900
                            Maximum theoretical value of tilt.
        :type max_raw_tilt: int
        :param max_degree_azimuth: OPTIONAL, DEFAULT = 360
                                   Maximum degree value of azimuth.
        :type max_degree_azimuth: int
        :param max_degree_tilt: OPTIONAL, DEFAULT = 90
                                Maximum degree value of tilt.
        :type max_degree_tilt: int
        :param max_pressure: OPTIONAL, DEFAULT = 32767
                             Maximum theoretical value of pressure.
        :type max_pressure: int
        :param pressure_levels: OPTIONAL, DEFAULT = 8192
                                Level of pressures of the device.
        :type pressure_levels: int
        :param angles_to_degrees: OPTIONAL, DEFAULT = True
                                  Transform angles to degrees
        :type angles_to_degrees: bool
        :param shift_to_zero: OPTIONAL, DEFAULT = True
                              Shift axis values to start from 0,0 coordinates
        :type shift_to_zero: bool
        """
        self.transformer.transform_all_units(
            self,
            conversion_type=conversion_type,
            lpi_value=lpi_value,
            lpmm_value=lpmm_value,
            max_raw_azimuth=max_raw_azimuth,
            max_raw_tilt=max_raw_tilt,
            max_degree_azimuth=max_degree_azimuth,
            max_degree_tilt=max_degree_tilt,
            max_pressure=max_pressure,
            pressure_levels=pressure_levels,
            angles_to_degrees=angles_to_degrees,
            shift_to_zero=shift_to_zero)

    def transform_axis_to_mm(
            self,
            conversion_type=transformer.LPI,
            lpi_value=transformer.LPI_VALUE,
            lpmm_value=transformer.LPMM_VALUE,
            shift_to_zero=True):

        """
        Transforms X,Y axis to millimeters.

        :param conversion_type: OPTIONAL ["lpi"|"lpmm"|"mm"], DEFAULT="lpi".
                                Set the capturing method used for mapping;
                                "lpi" for inch; "lpmm" for millimeters;
                                "mm" for direct to millimeters
        :type conversion_type: str
        :param lpi_value: OPTIONAL, DEFAULT = 5080
                          Set lpi value of digitizing tablet
        :type lpi_value: int
        :param lpmm_value: OPTIONAL, DEFAULT = 200
                           Set lpmm value of digitizing tablet
        :type lpmm_value: int
        :param shift_to_zero: OPTIONAL, DEFAULT = True
                              Shift axis values to start from 0,0 coordinates
        :type shift_to_zero: bool
        """

        self.transformer.transform_axis(self, conversion_type=conversion_type, lpi_value=lpi_value,
                                        lpmm_value=lpmm_value, shift_to_zero=shift_to_zero)

    def transform_time_to_seconds(self):
        """ Transform time to seconds """
        self.time = self.transformer.transform_time_to_seconds(self.time)

    def normalize_pressure(
            self,
            max_pressure=transformer.MAX_PRESSURE_VALUE,
            pressure_levels=transformer.PRESSURE_LEVELS):
        """
        Normalizes pressure to pressure level of the device.

        :param max_pressure: OPTIONAL, DEFAULT = 32767
                             max theoretical raw pressure value
        :type max_pressure: int
        :param pressure_levels: OPTIONAL, DEFAULT = 8192
                                level of pressure of the device
        :type pressure_levels: int
        """

        self.pressure = self.transformer.normalize_pressure(self.pressure,
                                                            max_value=max_pressure,
                                                            pressure_levels=pressure_levels)

    def transform_angle_to_degree(self, angle=None, max_raw_value=None, max_degree_value=None):
        """
        Transforms raw angle to degrees.

        :param angle: Angle that should bne converted [tilt, azimuth]
        :type angle: str
        :param max_raw_value: OPTIONAL, Maximal theoretical value of raw angle
        :type max_raw_value: int
        :param max_degree_value: OPTIONAL,  Maximal value of angle in degrees
        :type max_degree_value: int
        """

        # For tilt
        if angle == self.TILT:
            self.tilt = self.transformer.transform_angle(
                self.tilt,
                max_raw_value=max_raw_value or self.transformer.MAX_TILT_VALUE,
                max_degree_value=max_degree_value or self.transformer.MAX_TILT_DEGREE)

        # For Azimuth
        elif angle == self.AZIMUTH:
            self.azimuth = self.transformer.transform_angle(
                self.azimuth,
                max_raw_value=max_raw_value or self.transformer.MAX_AZIMUTH_VALUE,
                max_degree_value=max_degree_value or self.transformer.MAX_AZIMUTH_DEGREE)

        else:
            raise TransformerAngleTypeException(angle)

    # ---------------------- #
    # Meta data manipulation #
    # ---------------------- #

    def add_meta_data(self, meta_data):
        """Adds meta data to the HandwritingSample object from dictionary"""
        self.meta.update({"updated_on": datetime.utcnow().strftime(self.DATE_FORMAT)})
        self.meta.update({**meta_data})

    # -------------- #
    # Visualisation  #
    # -------------- #

    def plot_on_surface(self, x_label=None, y_label=None, save_path=None):
        """
        Plot on surface data

        :param x_label: OPTIONAL, label of X axis
        :type x_label: str
        :param y_label: OPTIONAL, label of Y axis
        :type y_label: str
        :param save_path: OPTIONAL, set save path if you wish to save the figure
        :type save_path: str

        :return: axis and plot objects
        """
        return self.visualizer.plot_on_surface_movement(self, x_label=x_label, y_label=y_label, save_as=save_path)

    def plot_in_air(self, x_label=None, y_label=None, save_path=None):
        """
        Plot in air data

        :param x_label: OPTIONAL, label of X axis
        :type x_label: str
        :param y_label: OPTIONAL, label of Y axis
        :type y_label: str
        :param save_path: OPTIONAL, set save path if you wish to save the figure
        :type save_path: str

        :return: axis and plot objects
        """
        return self.visualizer.plot_in_air_movement(self, x_label=x_label, y_label=y_label, save_as=save_path)

    def plot_separate_movements(self, x_label=None, y_label=None, save_path=None):
        """
        Plot separate movement in one plot (on_surface + in_air)

        :param x_label: OPTIONAL, label of X axis
        :type x_label: str
        :param y_label: OPTIONAL, label of Y axis
        :type y_label: str
        :param save_path: OPTIONAL, set save path if you wish to save the figure
        :type save_path: str

        :return: axis and plot objects
        """
        return self.visualizer.plot_separate_movements(self, x_label=x_label, y_label=y_label, save_as=save_path)

    def plot_strokes(self, x_label=None, y_label=None, save_path=None):
        """
        Plot separate strokes in one plot

        :param x_label: OPTIONAL, label of X axis
        :type x_label: str
        :param y_label: OPTIONAL, label of Y axis
        :type y_label: str
        :param save_path: OPTIONAL, set save path if you wish to save the figure
        :type save_path: str

        :return: axis and plot objects
        """
        return self.visualizer.plot_strokes(self, x_label=x_label, y_label=y_label, save_as=save_path)

    def plot_all_data(self, x_label=None, save_path=None):
        """
        Plot individual plots for each data attribute (x,y,time,azimuth,tilt,pressure)

        :param x_label: OPTIONAL, label of X axis
        :type x_label: str
        :param save_path: OPTIONAL, set save path if you wish to save the figure
        :type save_path: str
        """
        return self.visualizer.plot_all_modalities(self, x_label=x_label, save_as=save_path)



import os
import time
from datetime import datetime


def log(text, topic, be_verbose=False, is_verbose=False):
    """
    Logs the specified text and topic.

    More info:
    ``be_verbose`` is True if logging should be verbose
    ``is_verbose`` is True if the specified text to log is verbose

    :param text: a text to log
    :type text: str
    :param topic: a topic to log
    :type topic: str
    :param be_verbose: verbosity of the logging, defaults to False
    :type be_verbose: bool, optional
    :param is_verbose: verbosity of the text, defaults to False
    :type is_verbose: bool, optional
    :return: None
    :rtype: None type
    """
    if not is_verbose or (be_verbose and is_verbose):
        print(
            f"{datetime.fromtimestamp(time.time()).strftime('%Y-%m-%d %H:%M:%S')} - "
            f"{os.getppid()} - "
            f"{topic} - "
            f"{text}")



import pathlib
from setuptools import setup, find_packages


# Get the long description from the README file
long_description = (pathlib.Path(__file__).parent.resolve() / "README.md").read_text(encoding="utf-8")

# Prepare the packages and requirements
packages = find_packages(where="src")
requires = [
    "numpy",
    "pandas",
    "matplotlib"
]

# Prepare the setup
setup(
    name="handwriting-sample",
    version="1.0.1",
    description="Handwriting sample",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/BDALab/handwriting-sample",
    author="Brain Diseases Analysis Laboratory",
    author_email="mucha@vut.cz",
    packages=packages,
    package_data={"": ["LICENSE"]},
    package_dir={"": "src"},
    include_package_data=True,
    install_requires=requires,
    python_requires=">=3.7",
    license="MIT",
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: Implementation :: PyPy",
        "Operating System :: OS Independent"
    ]
)
